from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import orm
from vortex.Tuple import Tuple
from vortex.Tuple import addTupleType

from peek_plugin_livedb._private.PluginNames import livedbTuplePrefix
from .DeclarativeBase import DeclarativeBase


@addTupleType
class LiveDbModelSet(DeclarativeBase, Tuple):
    __tablename__ = "LiveDbModelSet"
    __tupleType__ = livedbTuplePrefix + __tablename__

    id = Column(Integer, primary_key=True, autoincrement=True)
    key = Column(String, nullable=False, unique=True)
    name = Column(String, nullable=False, unique=True)
    comment = Column(String)

    propsJson = Column(String)

    @orm.reconstructor
    def __init__(self, **kwargs):
        Tuple.__init__(self, **kwargs)


def getOrCreateLiveDbModelSet(session, modelSetKey: str) -> LiveDbModelSet:
    qry = session.query(LiveDbModelSet).filter(
        LiveDbModelSet.key == modelSetKey
    )
    if not qry.count():
        session.add(LiveDbModelSet(key=modelSetKey, name=modelSetKey))
        session.commit()

    return qry.one()



import logging
from vortex.sqla_orm.OrmCrudHandler import OrmCrudHandler

from peek_plugin_livedb._private.PluginNames import livedbFilt
from peek_plugin_livedb._private.storage.Setting import (
    SettingProperty,
    globalSetting,
)

logger = logging.getLogger(__name__)

# This dict matches the definition in the Admin angular app.
filtKey = {"key": "admin.Edit.SettingProperty"}
filtKey.update(livedbFilt)


# This is the CRUD handler
class __CrudHandler(OrmCrudHandler):
    # The UI only edits the global settings
    # You could get more complicated and have the UI edit different groups of settings.
    def createDeclarative(self, session, payloadFilt):
        return [p for p in globalSetting(session).propertyObjects]


# This method creates an instance of the handler class.
def makeSettingPropertyHandler(dbSessionCreator):
    handler = __CrudHandler(
        dbSessionCreator, SettingProperty, filtKey, retreiveAll=True
    )

    logger.debug("Started")
    return handler



from typing import List

from abc import ABCMeta, abstractmethod
from twisted.internet.defer import Deferred

from peek_plugin_livedb.tuples.ImportLiveDbItemTuple import (
    ImportLiveDbItemTuple,
)
from peek_plugin_livedb.tuples.LiveDbDisplayValueUpdateTuple import (
    LiveDbDisplayValueUpdateTuple,
)
from peek_plugin_livedb.tuples.LiveDbRawValueUpdateTuple import (
    LiveDbRawValueUpdateTuple,
)


class LiveDBWriteApiABC(metaclass=ABCMeta):
    @abstractmethod
    def updateRawValues(
        self, modelSetName: str, updates: List[LiveDbRawValueUpdateTuple]
    ) -> Deferred:
        """Process Live DB Raw Value Updates

        Tells the live db that values have updated in the field, or wherever.

        :param modelSetName:  The name of the model set for the live db
        :param updates: A list of tuples containing the value updates

        :return: A deferred that fires when the update is complete.
        :rtype: bool

        """

    @abstractmethod
    def importLiveDbItems(
        self, modelSetName: str, newItems: List[ImportLiveDbItemTuple]
    ) -> Deferred:
        """Import LiveDB Items

        Create new Live DB Items with Raw + Display values

        If an item already exists, it's value is update.

        :param modelSetName:  The name of the model set for the live db
        :param newItems: A list of tuples containing the value updates

        :return: A deferred that fires when the inserts are complete.
        :rtype: bool

        """

    @abstractmethod
    def prioritiseLiveDbValueAcquisition(
        self, modelSetName: str, liveDbKeys: List[str]
    ) -> Deferred:
        """Prioritise LiveDB Value Acquisitions

        When this method was first created, it was used for the diagram to tell the
        RealTime agent which keys to update as they were viewed by the user.

        :param modelSetName:  The name of the model set for the live db
        :param liveDbKeys: A list of the livedb keys to watch

        :return: A deferred that fires with True
        :rtype: bool

        """

    @abstractmethod
    def pollLiveDbValueAcquisition(
        self, modelSetName: str, liveDbKeys: List[str]
    ) -> Deferred:
        """Poll LiveDB Value Acquisitions

        Tell the LiveDB plugin the liveDbKeys must be polled.

        This method should be called after a re-import of an existing grid.

        :param modelSetName:  The name of the model set for the live db
        :param liveDbKeys: A list of the livedb keys to poll

        :return: A deferred that fires with True
        :rtype: bool

        """



import logging
from typing import List

from peek_abstract_chunked_index.private.server.controller.ACIProcessorQueueControllerABC import (
    ACIProcessorQueueControllerABC,
    ACIProcessorQueueBlockItem,
)
from peek_abstract_chunked_index.private.server.controller.ACIProcessorStatusNotifierABC import (
    ACIProcessorStatusNotifierABC,
)
from peek_abstract_chunked_index.private.tuples.ACIProcessorQueueTupleABC import (
    ACIProcessorQueueTupleABC,
)
from peek_plugin_livedb._private.server.controller.AdminStatusController import (
    AdminStatusController,
)
from peek_plugin_livedb._private.storage.LiveDbItem import LiveDbItem
from peek_plugin_livedb._private.storage.LiveDbModelSet import (
    getOrCreateLiveDbModelSet,
)
from peek_plugin_livedb._private.storage.LiveDbRawValueQueue import (
    LiveDbRawValueQueue,
)
from peek_plugin_livedb.tuples.LiveDbRawValueUpdateTuple import (
    LiveDbRawValueUpdateTuple,
)
from vortex.DeferUtil import deferToThreadWrapWithLogger

logger = logging.getLogger(__name__)


class _Notifier(ACIProcessorStatusNotifierABC):
    def __init__(self, adminStatusController: AdminStatusController):
        self._adminStatusController = adminStatusController

    def setProcessorStatus(self, state: bool, queueSize: int):
        self._adminStatusController.status.rawValueQueueStatus = state
        self._adminStatusController.status.rawValueQueueSize = queueSize
        self._adminStatusController.notify()

    def addToProcessorTotal(self, delta: int):
        self._adminStatusController.status.rawValueProcessedTotal += delta
        self._adminStatusController.notify()

    def setProcessorError(self, error: str):
        self._adminStatusController.status.rawValueLastError = error
        self._adminStatusController.notify()


class LiveDbValueUpdateQueueController(ACIProcessorQueueControllerABC):
    # Prioritize the livedb updater.
    MAX_CPU_PERCENTAGE = 85.00

    QUEUE_ITEMS_PER_TASK = 500
    POLL_PERIOD_SECONDS = 0.200

    QUEUE_BLOCKS_MAX = 20
    QUEUE_BLOCKS_MIN = 4

    WORKER_TASK_TIMEOUT = 60.0

    _logger = logger
    _QueueDeclarative: ACIProcessorQueueTupleABC = LiveDbRawValueQueue
    _VacuumDeclaratives = (LiveDbRawValueQueue, LiveDbItem)

    def __init__(
        self, ormSessionCreator, adminStatusController: AdminStatusController
    ):
        ACIProcessorQueueControllerABC.__init__(
            self, ormSessionCreator, _Notifier(adminStatusController)
        )

        self._modelSetIdByKey = {}

    def _sendToWorker(self, block: ACIProcessorQueueBlockItem):
        from peek_plugin_livedb._private.worker.tasks.LiveDbItemUpdateTask import (
            updateValues,
        )

        return updateValues.delay(block.itemsEncodedPayload)

    def _processWorkerResults(self, results):
        pass

    # ---------------
    # Deduplicate method

    def _dedupeQueueSql(self, lastFetchedId: int, dedupeLimit: int):
        """Deduplicate Queue

        NOTE: In this SQL, we take the last value added to the queue, as it will have
        the latest value.

        """
        return """
                 with sq_raw as (
                    SELECT "id", "modelSetId", "key"
                    FROM pl_livedb."LiveDbRawValueQueue"
                    WHERE id > %(id)s
                    LIMIT %(limit)s
                ), sq as (
                    -- Select the latest queued update, this does require that updates
                    -- come in, in order
                    SELECT max(id) as "maxId", "modelSetId", "key"
                    FROM sq_raw
                    GROUP BY  "modelSetId", "key"
                    HAVING count("key") > 1
                )
                DELETE
                FROM pl_livedb."LiveDbRawValueQueue"
                     USING sq sq1
                WHERE pl_livedb."LiveDbRawValueQueue"."id" != sq1."maxId"
                    AND pl_livedb."LiveDbRawValueQueue"."id" > %(id)s
                    AND pl_livedb."LiveDbRawValueQueue"."modelSetId" = sq1."modelSetId"
                    AND pl_livedb."LiveDbRawValueQueue"."key" = sq1."key"
                    
            """ % {
            "id": lastFetchedId,
            "limit": dedupeLimit,
        }

    # ---------------
    # Insert into Queue methods

    @deferToThreadWrapWithLogger(logger)
    def queueData(
        self, modelSetKey: str, updates: List[LiveDbRawValueUpdateTuple]
    ):
        if not updates:
            return

        ormSession = self._dbSessionCreator()
        try:
            logger.debug("Queueing %s raw values for compile", len(updates))

            if modelSetKey not in self._modelSetIdByKey:
                modelSet = getOrCreateLiveDbModelSet(
                    ormSession, modelSetKey=modelSetKey
                )
                self._modelSetIdByKey[modelSet.key] = modelSet.id

            modelSetId = self._modelSetIdByKey[modelSetKey]

            inserts = []
            for update in updates:
                inserts.append(
                    dict(
                        modelSetId=modelSetId,
                        key=update.key,
                        rawValue=update.rawValue,
                    )
                )

            ormSession.execute(LiveDbRawValueQueue.__table__.insert(), inserts)
            ormSession.commit()

        finally:
            ormSession.close()



from abc import ABCMeta, abstractmethod
from typing import Optional, List

from reactivex.subject import Subject
from twisted.internet.defer import Deferred


class LiveDBReadApiABC(metaclass=ABCMeta):
    @abstractmethod
    def priorityKeysObservable(self, modelSetName: str) -> Subject:
        """Priority Live DB Key Observable

        This observable emits list of keys that the live db acquisition plugins should
        prioritised, these keys will be monitored until the next
        priorityLiveDbKey update.

        This list will represent keys relating to the objects that are
        currently being viewed.

        :param modelSetName:  The name of the model set of the keys to observe.

        :return: An observable that emits a List[str].

        """

    @abstractmethod
    def pollKeysObservable(self, modelSetName: str) -> Subject:
        """Poll Live DB Key Observable

        This observable emits list of keys that the live db acquisition plugins should
        poll ONCE.

        This list will represent keys relating to the objects that are
        currently being viewed.

        :param modelSetName:  The name of the model set of the keys to observe.

        :return: An observable that emits a List[str].

        """

    @abstractmethod
    def itemAdditionsObservable(self, modelSetName: str) -> Subject:
        """Live DB Tuple Added Items Observable

        Return an observable that fires when livedb items are added

        :param modelSetName: The name of the model set for the live db

        :return: An observable that fires when keys are removed from the live db
        :rtype: An observable that emits List[LiveDbDisplayValueTuple]

        """

    @abstractmethod
    def itemDeletionsObservable(self, modelSetName: str) -> Subject:
        """Live DB Tuple Removed Items Observable

        Return an observable that fires when livedb items are removed

        :param modelSetName:  The name of the model set for the live db

        :return: An observable that fires when keys are removed from the live db
        :rtype: An observable that emits List[str]

        """

    @abstractmethod
    def bulkLoadDeferredGenerator(
        self,
        modelSetName: str,
        keyList: Optional[List[str]] = None,
        chunkSize: int = 2500,
    ) -> Deferred:
        """Live DB Tuples

        Return a generator that returns deferreds that are fired with chunks of the
         entire live db.

        :param chunkSize: The number of items to return for each chunk
        :param modelSetName:  The name of the model set for the live db
        :param keyList:  An optional list of keys that the data is required for

        :return: A deferred that fires with a list of tuples
        :rtype: C{LiveDbDisplayValueTuple}

        This is served up in chunks to prevent ballooning the memory usage.

        Here is an example of how to use this method

        ::

                @inlineCallbacks
                def loadFromDiagramApi(diagramLiveDbApi:DiagramLiveDbApiABC):
                    deferredGenerator = diagramLiveDbApi.bulkLoadDeferredGenerator("modelName")

                    while True:
                        d = next(deferredGenerator)
                        liveDbValueTuples = yield d # List[LiveDbDisplayValueTuple]

                        # The end of the list is marked my an empty result
                        if not liveDbValueTuples:
                            break

                        # TODO, do something with this chunk of liveDbValueTuples



        """

    @abstractmethod
    def rawValueUpdatesObservable(self, modelSetName: str) -> Subject:
        """Raw Value Update Observable

        Return an observable that fires with lists of C{LiveDbRawValueTuple} tuples
        containing updates to live db values.

        :param modelSetName:  The name of the model set for the live db

        :return: An observable that fires when values are updated in the livedb
        :rtype: Subject[List[LiveDbRawValueTuple]]

        """

    @abstractmethod
    def displayValueUpdatesObservable(self, modelSetName: str) -> Subject:
        """Display Value Update Observable

        Return an observable that fires with lists of C{LiveDbDisplayValueTuple} tuples
        containing updates to live db values.

        :param modelSetName:  The name of the model set for the live db

        :return: An observable that fires when values are updated in the livedb
        :rtype: An observable that fires with List[LiveDbDisplayValueTuple]

        """



"""added queue table

Peek Plugin Database Migration Script

Revision ID: d4ffb9c5cb27
Revises: 19b2a8d326ad
Create Date: 2020-04-18 16:15:31.166637

"""

# revision identifiers, used by Alembic.
revision = "d4ffb9c5cb27"
down_revision = "19b2a8d326ad"
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "LiveDbRawValueQueue",
        sa.Column("id", sa.BigInteger(), autoincrement=True, nullable=False),
        sa.Column("modelSetId", sa.Integer(), nullable=False),
        sa.Column("key", sa.String(), nullable=False),
        sa.Column("rawValue", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id", "modelSetId", "key", "rawValue"),
        schema="pl_livedb",
    )

    op.add_column(
        "LiveDbModelSet",
        sa.Column("key", sa.String(), nullable=True),
        schema="pl_livedb",
    )

    op.execute(
        '''UPDATE "pl_livedb"."LiveDbModelSet"
                  SET "key" = "name"'''
    )

    op.alter_column(
        "LiveDbModelSet",
        "key",
        type_=sa.String(),
        nullable=False,
        schema="pl_livedb",
    )
    op.create_unique_constraint(
        "idx_liveDbModelSet_key", "LiveDbModelSet", ["key"], schema="pl_livedb"
    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(
        "idx_liveDbModelSet_key",
        "LiveDbModelSet",
        schema="pl_livedb",
        type_="unique",
    )
    op.drop_column("LiveDbModelSet", "key", schema="pl_livedb")
    op.drop_table("LiveDbRawValueQueue", schema="pl_livedb")
    # ### end Alembic commands ###



"""Removed string lengths

Peek Plugin Database Migration Script

Revision ID: 19b2a8d326ad
Revises: ab9da4532175
Create Date: 2019-06-10 21:52:38.648849

"""

# revision identifiers, used by Alembic.
revision = "19b2a8d326ad"
down_revision = "ab9da4532175"
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    op.alter_column(
        "LiveDbItem",
        "key",
        type_=sa.String(),
        nullable=False,
        schema="pl_livedb",
    )
    op.alter_column(
        "LiveDbItem",
        "rawValue",
        type_=sa.String(),
        nullable=True,
        schema="pl_livedb",
    )
    op.alter_column(
        "LiveDbItem",
        "displayValue",
        type_=sa.String(),
        nullable=True,
        schema="pl_livedb",
    )
    op.alter_column(
        "LiveDbItem",
        "importHash",
        type_=sa.String(),
        nullable=True,
        schema="pl_livedb",
    )
    op.alter_column(
        "LiveDbItem",
        "propsJson",
        type_=sa.String(),
        nullable=True,
        schema="pl_livedb",
    )

    op.alter_column(
        "LiveDbModelSet",
        "name",
        type_=sa.String(),
        nullable=False,
        schema="pl_livedb",
    )
    op.alter_column(
        "LiveDbModelSet",
        "propsJson",
        type_=sa.String(),
        nullable=True,
        schema="pl_livedb",
    )


def downgrade():
    pass



import logging

from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import orm
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import Index
from sqlalchemy.sql.schema import Sequence
from vortex.Tuple import JSON_EXCLUDE
from vortex.Tuple import Tuple
from vortex.Tuple import addTupleType

from peek_plugin_livedb._private.PluginNames import livedbTuplePrefix
from .DeclarativeBase import DeclarativeBase
from .LiveDbModelSet import LiveDbModelSet

logger = logging.getLogger(__name__)


@addTupleType
class LiveDbItem(DeclarativeBase, Tuple):
    __tupleTypeShort__ = "LDK"
    __tablename__ = "LiveDbItem"
    __tupleType__ = livedbTuplePrefix + __tablename__

    NUMBER_VALUE = 0
    STRING_VALUE = 1
    COLOR = 2
    LINE_WIDTH = 3
    LINE_STYLE = 4
    GROUP_PTR = 5

    id_seq = Sequence(
        "LiveDbItem_id_seq",
        metadata=DeclarativeBase.metadata,
        schema=DeclarativeBase.metadata.schema,
    )
    id = Column(
        Integer,
        id_seq,
        server_default=id_seq.next_value(),
        primary_key=True,
        autoincrement=False,
    )

    modelSetId = Column(
        Integer,
        ForeignKey("LiveDbModelSet.id", ondelete="CASCADE"),
        doc=JSON_EXCLUDE,
        nullable=False,
    )
    modelSet = relationship(LiveDbModelSet)

    # comment="The unique reference of the value we want from the live db"
    key = Column(String, nullable=False)

    # comment="The last value from the source"
    rawValue = Column(String)

    # comment="The PEEK value, converted to PEEK IDs if required (Color for example)"
    displayValue = Column(String)

    # comment="The type of data this value represents"
    dataType = Column(Integer, nullable=False)

    importHash = Column(String)

    # Store custom props for this link
    propsJson = Column(String)

    __table_args__ = (
        Index("idx_LiveDbDKey_importHash", importHash, unique=False),
        Index("idx_LiveDbDKey_modelSet_key", modelSetId, key, unique=True),
    )

    @orm.reconstructor
    def __init__(self, **kwargs):
        Tuple.__init__(self, **kwargs)



import logging
from datetime import datetime
from typing import List

import pytz
from sqlalchemy.sql.expression import select

from peek_plugin_base.storage.StorageUtil import makeCoreValuesSubqueryCondition
from peek_plugin_base.worker.task_db_conn import TaskDbConn
from peek_plugin_livedb._private.storage.LiveDbItem import LiveDbItem
from peek_plugin_livedb._private.storage.LiveDbModelSet import (
    getOrCreateLiveDbModelSet,
)
from peek_plugin_base.worker.task import addPeekWorkerTask
from peek_plugin_livedb.tuples.ImportLiveDbItemTuple import (
    ImportLiveDbItemTuple,
)

logger = logging.getLogger(__name__)


@addPeekWorkerTask()
def importLiveDbItems(
    modelSetKey: str, newItems: List[ImportLiveDbItemTuple]
) -> List[str]:
    """Compile Grids Task


    :param modelSetKey: The model set name
    :param newItems: The list of new items
    :returns: A list of grid keys that have been updated.
    """

    startTime = datetime.now(pytz.utc)

    session = TaskDbConn.getDbSession()
    engine = TaskDbConn.getDbEngine()
    conn = engine.connect()
    transaction = conn.begin()

    liveDbTable = LiveDbItem.__table__
    try:
        liveDbModelSet = getOrCreateLiveDbModelSet(session, modelSetKey)

        # This will remove duplicates
        itemsByKey = {i.key: i for i in newItems}

        allKeys = list(itemsByKey)
        existingKeys = set()

        # Query for existing keys, in 1000 chinks
        chunkSize = 1000
        offset = 0
        while True:
            chunk = allKeys[offset : offset + chunkSize]
            if not chunk:
                break
            offset += chunkSize
            stmt = (
                select(liveDbTable.c.key)
                .where(liveDbTable.c.modelSetId == liveDbModelSet.id)
                .where(
                    makeCoreValuesSubqueryCondition(
                        engine, liveDbTable.c.key, chunk
                    )
                )
            )

            result = conn.execute(stmt)

            existingKeys.update([o[0] for o in result.fetchall()])

        inserts = []
        newKeys = []

        for newItem in itemsByKey.values():
            if newItem.key in existingKeys:
                continue

            inserts.append(
                dict(
                    modelSetId=liveDbModelSet.id,
                    key=newItem.key,
                    dataType=newItem.dataType,
                    rawValue=newItem.rawValue,
                    displayValue=newItem.displayValue,
                    importHash=newItem.importHash,
                )
            )

            newKeys.append(newItem.key)

        if not inserts:
            return []

        conn.execute(LiveDbItem.__table__.insert(), inserts)

        transaction.commit()
        logger.info(
            "Inserted %s LiveDbItems, %s already existed, in %s",
            len(inserts),
            len(existingKeys),
            (datetime.now(pytz.utc) - startTime),
        )

        return newKeys

    except Exception as e:
        transaction.rollback()
        logger.debug("Task failed, but it will retry. %s", e)
        raise

    finally:
        conn.close()
        session.close()



import logging
from typing import Optional

from sqlalchemy import Column, BigInteger, Index
from sqlalchemy import Integer, String
from vortex.Tuple import Tuple, addTupleType

from peek_abstract_chunked_index.private.tuples.ACIProcessorQueueTupleABC import (
    ACIProcessorQueueTupleABC,
)
from .DeclarativeBase import DeclarativeBase
from ..PluginNames import livedbTuplePrefix

logger = logging.getLogger(__name__)


@addTupleType
class LiveDbRawValueQueue(Tuple, DeclarativeBase, ACIProcessorQueueTupleABC):
    __tablename__ = "LiveDbRawValueQueue"
    __tupleType__ = livedbTuplePrefix + __tablename__

    id = Column(BigInteger, primary_key=True, autoincrement=True)

    modelSetId = Column(Integer, nullable=False)
    key = Column(String, nullable=False)
    rawValue = Column(String)

    @classmethod
    def sqlCoreLoad(cls, row):
        return LiveDbRawValueQueue(
            id=row.id,
            modelSetId=row.modelSetId,
            key=row.key,
            rawValue=row.rawValue,
        )

    @property
    def ckiUniqueKey(self):
        """See LiveDbRawValueQueueTuple.ckiUniqueKey"""
        return "%s:%s" % (self.modelSetId, self.key)

    __table_args__ = (
        Index(
            "idx_LiveDbRawValueQueue_all",
            id,
            modelSetId,
            key,
            rawValue,
            unique=False,
        ),
    )



"""Initial

Peek Plugin Database Migration Script

Revision ID: 64309a77b48d
Revises: 
Create Date: 2017-07-13 16:22:26.973631

"""

# revision identifiers, used by Alembic.
revision = "64309a77b48d"
down_revision = None
branch_labels = None
depends_on = None

from sqlalchemy.dialects.mssql.base import MSDialect
from sqlalchemy.dialects.postgresql.base import PGDialect
from sqlalchemy.schema import Sequence, CreateSequence
from alembic import op
import sqlalchemy as sa
import geoalchemy2


def __isMssqlDialect():
    return isinstance(op.get_bind().engine.dialect, MSDialect)


def __isPostGreSQLDialect():
    return isinstance(op.get_bind().engine.dialect, PGDialect)


def __nextval(seqName):
    if __isMssqlDialect():
        return sa.text('(NEXT VALUE FOR "pl_livedb"."%s")' % seqName)

    return sa.text("nextval('pl_livedb.\"%s\"')" % seqName)


def upgrade():
    op.execute(
        CreateSequence(
            Sequence("LiveDbItem_id_seq", minvalue=0, schema="pl_livedb")
        )
    )

    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "LiveDbModelSet",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=50), nullable=False),
        sa.Column("comment", sa.String(), nullable=True),
        sa.Column("propsJson", sa.String(length=500), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        schema="pl_livedb",
    )
    op.create_table(
        "Setting",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=50), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        schema="pl_livedb",
    )

    op.create_table(
        "LiveDbItem",
        sa.Column(
            "id",
            sa.Integer(),
            server_default=__nextval("LiveDbItem_id_seq"),
            nullable=False,
        ),
        sa.Column("modelSetId", sa.Integer(), nullable=False),
        sa.Column("key", sa.String(length=50), nullable=False),
        sa.Column("rawValue", sa.String(length=255), nullable=True),
        sa.Column("displayValue", sa.String(length=255), nullable=True),
        sa.Column("dataType", sa.Integer(), nullable=False),
        sa.Column("importHash", sa.String(length=100), nullable=True),
        sa.Column("propsJson", sa.String(length=500), nullable=True),
        sa.ForeignKeyConstraint(
            ["modelSetId"], ["pl_livedb.LiveDbModelSet.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="pl_livedb",
    )
    op.create_index(
        "idx_LiveDbDKey_importHash",
        "LiveDbItem",
        ["importHash"],
        unique=False,
        schema="pl_livedb",
    )
    op.create_index(
        "idx_LiveDbDKey_liveDbKey",
        "LiveDbItem",
        ["key"],
        unique=False,
        schema="pl_livedb",
    )
    op.create_index(
        "idx_LiveDbDKey_modelSetId",
        "LiveDbItem",
        ["modelSetId"],
        unique=False,
        schema="pl_livedb",
    )
    op.create_table(
        "SettingProperty",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("settingId", sa.Integer(), nullable=False),
        sa.Column("key", sa.String(length=50), nullable=False),
        sa.Column("type", sa.String(length=16), nullable=True),
        sa.Column("int_value", sa.Integer(), nullable=True),
        sa.Column("char_value", sa.String(length=50), nullable=True),
        sa.Column("boolean_value", sa.Boolean(), nullable=True),
        sa.ForeignKeyConstraint(
            ["settingId"],
            ["pl_livedb.Setting.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="pl_livedb",
    )
    op.create_index(
        "idx_SettingProperty_settingId",
        "SettingProperty",
        ["settingId"],
        unique=False,
        schema="pl_livedb",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        "idx_SettingProperty_settingId",
        table_name="SettingProperty",
        schema="pl_livedb",
    )
    op.drop_table("SettingProperty", schema="pl_livedb")
    op.drop_index(
        "idx_LiveDbDKey_modelSetId", table_name="LiveDbItem", schema="pl_livedb"
    )
    op.drop_index(
        "idx_LiveDbDKey_liveDbKey", table_name="LiveDbItem", schema="pl_livedb"
    )
    op.drop_index(
        "idx_LiveDbDKey_importHash", table_name="LiveDbItem", schema="pl_livedb"
    )
    op.drop_table("LiveDbItem", schema="pl_livedb")
    op.drop_table("Setting", schema="pl_livedb")
    op.drop_table("LiveDbModelSet", schema="pl_livedb")
    # ### end Alembic commands ###



from typing import Type

from peek_plugin_base.agent.PluginAgentEntryHookABC import (
    PluginAgentEntryHookABC,
)
from peek_plugin_base.client.PluginClientEntryHookABC import (
    PluginClientEntryHookABC,
)
from peek_plugin_base.server.PluginLogicEntryHookABC import (
    PluginLogicEntryHookABC,
)
from peek_plugin_base.worker.plugin_worker_entry_hook_abc import (
    PluginWorkerEntryHookABC,
)

__version__ = '5.0.0'


def peekLogicEntryHook() -> Type[PluginLogicEntryHookABC]:
    from ._private.server.LogicEntryHook import LogicEntryHook

    return LogicEntryHook


def peekAgentEntryHook() -> Type[PluginAgentEntryHookABC]:
    from ._private.agent.AgentEntryHook import AgentEntryHook

    return AgentEntryHook


def peekOfficeEntryHook() -> Type[PluginClientEntryHookABC]:
    from ._private.client.ClientEntryHook import ClientEntryHook

    return ClientEntryHook


def peekFieldEntryHook() -> Type[PluginClientEntryHookABC]:
    from ._private.client.ClientEntryHook import ClientEntryHook

    return ClientEntryHook


def peekWorkerEntryHook() -> Type[PluginWorkerEntryHookABC]:
    from ._private.worker.WorkerEntryHook import WorkerEntryHook

    return WorkerEntryHook



import logging
from typing import List

from twisted.internet.defer import Deferred, inlineCallbacks
from vortex.Payload import Payload

from peek_plugin_livedb._private.server.LiveDBReadApi import LiveDBReadApi
from peek_plugin_livedb._private.worker.tasks.LiveDbItemImportTask import (
    importLiveDbItems,
)
from peek_plugin_livedb.tuples.ImportLiveDbItemTuple import (
    ImportLiveDbItemTuple,
)

logger = logging.getLogger(__name__)


class LiveDbImportController:
    """LiveDB Import Controller"""

    def __init__(self, dbSessionCreator):
        self._dbSessionCreator = dbSessionCreator

    def setReadApi(self, readApi: LiveDBReadApi):
        self._readApi = readApi

    def shutdown(self):
        self._readApi = None

    @inlineCallbacks
    def importLiveDbItems(
        self, modelSetKey: str, newItems: List[ImportLiveDbItemTuple]
    ) -> Deferred:
        """Import Live DB Items

        1) set the  coordSetId

        2) Drop all disps with matching importGroupHash

        :param modelSetKey: The name of the model set
        :param newItems: The items to add or update to the live db
        :return:
        """

        newKeys = yield importLiveDbItems.delay(
            modelSetKey=modelSetKey, newItems=newItems
        )

        newTuples = []

        deferredGenerator = self._readApi.bulkLoadDeferredGenerator(
            modelSetKey, keyList=newKeys
        )
        while True:
            d = next(deferredGenerator)
            result = yield d  # List[LiveDbDisplayValueTuple]

            # The end of the list is marked my an empty result
            if not result or not result.count:
                break

            payload = yield Payload().fromEncodedPayloadDefer(
                result.encodedPayload
            )

            newTuples += payload.tuples

        # If there are no tuples, do nothing
        if not newTuples:
            return

        # Notify the agent of the new keys.
        self._readApi.itemAdditionsObservable(modelSetKey).on_next(newTuples)



import logging
from typing import List

from sqlalchemy import select

from peek_plugin_base.worker.task_db_conn import TaskDbConn
from peek_plugin_livedb._private.storage.LiveDbItem import LiveDbItem
from peek_plugin_base.worker.task import addPeekWorkerTask
from peek_plugin_livedb.tuples.LiveDbDisplayValueTuple import (
    LiveDbDisplayValueTuple,
)

logger = logging.getLogger(__name__)


@addPeekWorkerTask()
def qryChunkInWorker(offset, limit) -> List[LiveDbDisplayValueTuple]:
    """Query Chunk

    This returns a chunk of LiveDB items from the database


    :param offset: The offset of the chunk
    :param limit: An encoded payload containing the updates
    :returns: List[LiveDbDisplayValueTuple] serialised in a payload json
    """

    table = LiveDbItem.__table__

    session = TaskDbConn.getDbSession()
    try:
        result = session.execute(
            select(
                table.c.key,
                table.c.dataType,
                table.c.rawValue,
                table.c.displayValue,
            )
            .order_by(table.c.id)
            .offset(offset)
            .limit(limit)
        )

        return [
            LiveDbDisplayValueTuple(
                key=o.key,
                dataType=o.dataType,
                rawValue=o.rawValue,
                displayValue=o.displayValue,
            )
            for o in result.fetchall()
        ]

    finally:
        session.close()



import logging
from collections import defaultdict
from datetime import datetime
from typing import List, Dict

import pytz
from sqlalchemy.sql.expression import bindparam, and_, select

from vortex.Payload import Payload

from peek_plugin_base.worker.task_db_conn import TaskDbConn
from peek_plugin_base.worker.task import addPeekWorkerTask
from peek_plugin_livedb._private.storage.LiveDbItem import LiveDbItem
from peek_plugin_livedb._private.storage.LiveDbModelSet import LiveDbModelSet
from peek_plugin_livedb._private.storage.LiveDbRawValueQueue import (
    LiveDbRawValueQueue,
)
from peek_plugin_livedb.tuples.LiveDbDisplayValueTuple import (
    LiveDbDisplayValueTuple,
)

logger = logging.getLogger(__name__)


@addPeekWorkerTask()
def updateValues(payloadEncodedArgs: bytes) -> None:
    """Compile Grids Task


    :param payloadEncodedArgs: The updates from the queue controller
    :returns: None
    """
    startTime = datetime.now(pytz.utc)

    argData = Payload().fromEncodedPayload(payloadEncodedArgs).tuples
    allModelUpdates: List[LiveDbRawValueQueue] = argData[0]
    queueItemIds = argData[1]

    # Group the data by model set
    updatesByModelSetId = defaultdict(list)
    for update in allModelUpdates:
        updatesByModelSetId[update.modelSetId].append(update)

    ormSession = TaskDbConn.getDbSession()
    try:
        for modelSetId, modelUpdates in updatesByModelSetId.items():
            _updateValuesForModelSet(modelSetId, modelUpdates, ormSession)

        # ---------------
        # delete the queue items
        dispQueueTable = LiveDbRawValueQueue.__table__
        ormSession.execute(
            dispQueueTable.delete().where(dispQueueTable.c.id.in_(queueItemIds))
        )

        ormSession.commit()

        # ---------------
        # Finally, tell log some statistics
        logger.info(
            "Updated %s raw values in %s",
            len(allModelUpdates),
            (datetime.now(pytz.utc) - startTime),
        )

    except Exception as e:
        logger.exception(e)
        raise

    finally:
        ormSession.close()


def _updateValuesForModelSet(modelSetId, modelUpdates, ormSession):
    # Try to load the Diagram plugins API
    try:
        from peek_plugin_diagram.worker.WorkerApi import (
            WorkerApi as DiagramWorkerApi,
        )
    except ImportError:
        logger.warning("Failed to load the diagram WorkerAPI")
        DiagramWorkerApi = None

    table = LiveDbItem.__table__

    # Load the Model Set
    liveDbModelSet = (
        ormSession.query(LiveDbModelSet)
        .filter(LiveDbModelSet.id == modelSetId)
        .one()
    )

    # Create a list of keys
    updatedKeys = [i.key for i in modelUpdates]

    # ---------------
    # Make a list of display items from the provided data
    displayItems = _makeDisplayValueTuples(
        liveDbModelSet, modelUpdates, ormSession, updatedKeys
    )

    # ---------------
    # Get the Diagram plugin to convert the live db values
    if DiagramWorkerApi:
        DiagramWorkerApi.updateLiveDbDisplayValues(
            ormSession,
            modelSetKey=liveDbModelSet.name,
            liveDbRawValues=displayItems,
        )

    # ---------------
    # Update the the values in the tables
    sql = (
        table.update()
        .where(
            and_(
                table.c.key == bindparam("b_key"),
                table.c.modelSetId == liveDbModelSet.id,
            )
        )
        .values(
            {
                "rawValue": bindparam("b_rawValue"),
                "displayValue": bindparam("b_displayValue"),
            }
        )
    )

    ormSession.execute(
        sql,
        [
            dict(
                b_key=o.key,
                b_rawValue=o.rawValue,
                b_displayValue=o.displayValue,
            )
            for o in displayItems
        ],
    )

    # ---------------
    # Tell the diagram plugin that livedb values have been updated
    if DiagramWorkerApi:
        DiagramWorkerApi.liveDbDisplayValueUpdateNotify(
            ormSession, modelSetKey=liveDbModelSet.key, updatedKeys=updatedKeys
        )


def _makeDisplayValueTuples(
    liveDbModelSet, modelUpdates, ormSession, updatedKeys
):
    # Load the key typ lookups
    dataTypeLookup = _getLiveDbKeyDatatypeDict(
        ormSession, liveDbModelSet, liveDbKeys=updatedKeys
    )

    displayItems = []
    for update in modelUpdates:
        displayItems.append(
            LiveDbDisplayValueTuple(
                key=update.key,
                dataType=dataTypeLookup.get(update.key),
                rawValue=update.rawValue,
            )
        )

    return displayItems


def _getLiveDbKeyDatatypeDict(
    ormSession, liveDbModelSet: LiveDbModelSet, liveDbKeys: List[str]
) -> Dict[str, int]:
    """Get Live DB DataTypes

    Return an array of items representing the display values from the LiveDB.

    :param ormSession: The SQLAlchemy orm session from the calling code.
    :param liveDbModelSet: The name of the model set to get the keys for
    :param liveDbKeys: An array of LiveDb Keys.

    :returns: An array of tuples.
    """
    liveDbTable = LiveDbItem.__table__
    modelTable = LiveDbModelSet.__table__

    if not liveDbKeys:
        return {}

    liveDbKeys = list(set(liveDbKeys))  # Remove duplicates if any exist.
    stmt = (
        select(liveDbTable.c.key, liveDbTable.c.dataType)
        .select_from(
            liveDbTable.join(
                modelTable, liveDbTable.c.modelSetId == modelTable.c.id
            )
        )
        .where(modelTable.c.name == liveDbModelSet.name)
        .where(liveDbTable.c.key.in_(liveDbKeys))
    )

    resultSet = ormSession.execute(stmt)
    return dict(resultSet.fetchall())



from peek_plugin_livedb._private.PluginNames import livedbTuplePrefix

from peek_plugin_livedb._private.storage.LiveDbItem import LiveDbItem
from vortex.Tuple import Tuple, addTupleType


@addTupleType
class ImportLiveDbItemTuple(Tuple):
    """Live DB Display Value Tuple

    This tuple stores a value of a key in the Live DB database.

    """

    __tupleType__ = livedbTuplePrefix + "ImportLiveDbItemTuple"
    __slots__ = ("key", "dataType", "rawValue", "displayValue", "importHash")

    DATA_TYPE_NUMBER_VALUE = LiveDbItem.NUMBER_VALUE
    DATA_TYPE_STRING_VALUE = LiveDbItem.STRING_VALUE
    DATA_TYPE_COLOR = LiveDbItem.COLOR
    DATA_TYPE_LINE_WIDTH = LiveDbItem.LINE_WIDTH
    DATA_TYPE_LINE_STYLE = LiveDbItem.LINE_STYLE
    DATA_TYPE_GROUP_PTR = LiveDbItem.GROUP_PTR



from peek_plugin_livedb._private.tuples.AdminStatusTuple import AdminStatusTuple
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

from peek_plugin_livedb._private.PluginNames import livedbFilt
from peek_plugin_livedb._private.PluginNames import livedbObservableName
from peek_plugin_livedb._private.server.controller.AdminStatusController import (
    AdminStatusController,
)
from peek_plugin_livedb._private.server.tuple_providers.AdminStatusTupleProvider import (
    AdminStatusTupleProvider,
)


def makeTupleDataObservableHandler(
    ormSessionCreator, adminStatusController: AdminStatusController
):
    """ " Make Tuple Data Observable Handler

    This method creates the observable object, registers the tuple providers and then
    returns it.

    :param adminStatusController:
    :param ormSessionCreator: A function that returns a SQLAlchemy session when called

    :return: An instance of :code:`TupleDataObservableHandler`

    """
    tupleObservable = TupleDataObservableHandler(
        observableName=livedbObservableName, additionalFilt=livedbFilt
    )

    # # Register TupleProviders here
    tupleObservable.addTupleProvider(
        AdminStatusTuple.tupleName(),
        AdminStatusTupleProvider(adminStatusController),
    )
    return tupleObservable



from typing import List, Dict

from sqlalchemy import select

from peek_plugin_base.storage.StorageUtil import (
    makeCoreValuesSubqueryCondition,
    makeOrmValuesSubqueryCondition,
)
from peek_plugin_livedb._private.storage.LiveDbItem import LiveDbItem
from peek_plugin_livedb._private.storage.LiveDbModelSet import (
    LiveDbModelSet,
    getOrCreateLiveDbModelSet,
)
from peek_plugin_livedb.tuples.LiveDbDisplayValueTuple import (
    LiveDbDisplayValueTuple,
)
from peek_plugin_livedb.tuples.LiveDbRawValueTuple import LiveDbRawValueTuple


class WorkerApi:
    """Worker Api

    This class allows other classes to work with the LiveDB plugin on the
    worker service.

    """

    _FETCH_SIZE = 5000

    @classmethod
    def getLiveDbDisplayValues(
        cls, ormSession, modelSetKey: str, liveDbKeys: List[str]
    ) -> List[LiveDbDisplayValueTuple]:
        """Get Live DB Display Values

        Return an array of items representing the display values from the LiveDB.

        :param ormSession: The SQLAlchemy orm session from the calling code.
        :param modelSetKey: The name of the model set to get the keys for
        :param liveDbKeys: An array of LiveDb Keys.

        :returns: An array of tuples.
        """
        if not liveDbKeys:
            return []

        liveDbModelSet = getOrCreateLiveDbModelSet(ormSession, modelSetKey)

        liveDbKeys = set(liveDbKeys)  # Remove duplicates if any exist.
        qry = (
            ormSession.query(LiveDbItem)
            .filter(LiveDbItem.modelSetId == liveDbModelSet.id)
            .filter(
                makeOrmValuesSubqueryCondition(
                    ormSession, LiveDbItem.key, list(liveDbKeys)
                )
            )
            .yield_per(cls._FETCH_SIZE)
        )

        results = []

        for item in qry:
            results.append(
                LiveDbDisplayValueTuple(
                    key=item.key,
                    displayValue=item.displayValue,
                    rawValue=item.rawValue,
                    dataType=item.dataType,
                )
            )

        return results
